// This code accompanies the blog post:
// http://www.darwenaudio.com/blog/using-dual-mcp23s17-chips-with-my-tiva-launchpad
//
// The following was written by Terence M. Darwen February 2017.  Some rights reserved.
//
// Contact Info:
// Email: tmdarwen@gmail.com
// Web: tmdarwen.com
//
// The below code is for using dual MCP23S17 I/O expander chips with a single SSI connection
// to the Tiva LaunchPad (TM4C123G).
//
// This is an extension of the a "Tutorial on SPI Communication" done by "AllAboutEE" on
// YouTube here: https://www.youtube.com/watch?v=VprofC5ssi4 
//
// Note that in the comments below I'll frequently mention "the datasheet".  I'm referring to Texas
// Instruments' "Tiva TM4C123GH6PM Microcontroller Data Sheet".  Currently, it can be found here:
// http://www.ti.com/lit/ds/spms376e/spms376e.pdf

#include <stdint.h>

// The base address of the System Control registers is 0x400F.E000.  A register map for this
// can be seen in the datasheet on page 232.
#define SYSTEM_CONTROL_BASE       0x400FE000
#define SYSTEM_CONTROL_RCGCGPIO   (SYSTEM_CONTROL_BASE + 0x608)
#define SYSTEM_CONTROL_RCGCSSI    (SYSTEM_CONTROL_BASE + 0x61C)

// The base address for GPIO Port B is 0x4000.5000.  Documentation starts on page 658.
#define GPIO_PORT_B_BASE          0x40005000
#define GPIO_PORT_B_DATA          (GPIO_PORT_B_BASE + 0x3FC)
#define GPIO_PORT_B_DIR           (GPIO_PORT_B_BASE + 0x400)
#define GPIO_PORT_B_AFSEL         (GPIO_PORT_B_BASE + 0x420)
#define GPIO_PORT_B_DEN           (GPIO_PORT_B_BASE + 0x51C)
#define GPIO_PORT_B_AMSEL         (GPIO_PORT_B_BASE + 0x528)
#define GPIO_PORT_B_PCTL          (GPIO_PORT_B_BASE + 0x52C)

// The base address for GPIO Port B is 0x4000.5000.  Documentation starts on page 658.
#define GPIO_PORT_B_BASE          0x40005000
#define GPIO_PORT_B_DATA          (GPIO_PORT_B_BASE + 0x3FC)   // "GPIO Data"
#define GPIO_PORT_B_GPIODIR       (GPIO_PORT_B_BASE + 0x400)   // "GPIO Direction"
#define GPIO_PORT_B_GPIOAFSEL     (GPIO_PORT_B_BASE + 0x420)   // "GPIO Alternate Function Select"
#define GPIO_PORT_B_GPIOODR       (GPIO_PORT_B_BASE + 0x50C)   // "GPIO Open Drain Select"
#define GPIO_PORT_B_GPIODEN       (GPIO_PORT_B_BASE + 0x51C)   // "GPIO Digital Enable"
#define GPIO_PORT_B_GPIOAMSEL     (GPIO_PORT_B_BASE + 0x528)   // "GPIO Analog Mode Select"
#define GPIO_PORT_B_GPIOPCTL      (GPIO_PORT_B_BASE + 0x52C)   // "GPIO Port Control"

// The base address for the SSI2 register is 0x4000.A000.  Documentation starts on page 967.
#define SSI2_BASE                 0x4000A000
#define SSI2_CR0                  (SSI2_BASE + 0x000)   // "SSI Control 1" pg 969
#define SSI2_CR1                  (SSI2_BASE + 0x004)   // "SSI Control 1" pg 971
#define SSI2_DR                   (SSI2_BASE + 0x008)   // "SSI Data" pg 973
#define SSI2_SR                   (SSI2_BASE + 0x00C)   // "SSI Data" pg 974
#define SSI2_CPSR                 (SSI2_BASE + 0x010)   // "SSI Clock Prescale" pg 976
#define SSI2_CC                   (SSI2_BASE + 0x0FC8)  // "SSI Clock Configuration" pg 984

// See page 12 of the Microchip MCP23S17 datasheet
#define MCP23S17_IO_DIR_A_REG     0x00
#define MCP23S17_IO_DIR_B_REG     0x01
#define MCP23S17_IOCON_REG        0x0A
#define MCP23S17_GPIO_A_DATA_REG  0x12
#define MCP23S17_GPIO_B_DATA_REG  0x13

// See page 15 of the Microchip MCP23S17 datasheet
#define MCP23S17_RWBIT_READ       0x01
#define MCP23S17_ALL_BITS_INPUT   0xFF   // See page 10: "When a bit is set, the corresponding pin becomes an input"
#define MCP23S17_CHIP1_ADDRESS    0x00
#define MCP23S17_CHIP2_ADDRESS    0x02
#define MCP23S17_FIXED_BITS       0x40
#define MCP23S17_HAEN             0x08   // "Hardware address enable"


void SetBits(uint32_t address, uint32_t bits)
{
	(*((volatile uint32_t*)address)) = bits;
}

void ClearBits(uint32_t address, uint32_t bits)
{
	(*((volatile uint32_t*)address)) &= (~bits);
}

void OrEqualBits(uint32_t address, uint32_t value)
{
	(*((volatile uint32_t*)address)) |= value;
}

void WaitUntilBitSet(uint32_t address, uint32_t bits)
{
	while(((*((volatile unsigned int*)address)) & bits) == 0) { };
}

uint8_t ReadByte(uint32_t address)
{
	return (*((volatile uint8_t*)address));
}


// In order for the SSI communication to work properly, we need to control the clock signal (pin B5) ourselves.
// This is because typical SSI communication sends a single byte of data and then the clock signal.  With this
// MCP23S17 chip we need to send 3 bytes between clock cycles.  For a good white board explanation of this,
// see https://www.youtube.com/watch?v=VprofC5ssi4
void ConfigureGPIOPinB5()
{
	// "AMSEL" is "GPIO Analog Mode Select" aka "GPIOAMSEL".  Docs on page 687.  As the documentation states,
	// clearing the bit for the pin disables analog function of the pin.  Note that 1<<5 corresponds to
	// 0010.0000 binary (i.e. pin 5).
	ClearBits(GPIO_PORT_B_AMSEL, 1<<5);

	// This is the "GPIO Port Control (GPIOPCTL)" register.  Docs are on page 688.  "The GPIOPCTL register is
	// used in conjunction with the GPIOAFSEL register and selects the specific peripheral signal for each GPIO
	// pin when using the alternate function mode".  Clearing the bit associated with PB5 specifies that
	// we will be using it as a general purpose I/O pin.
	ClearBits(GPIO_PORT_B_PCTL, 0xF0000);

	// This sets the direction of the PB5 pin to output.  Docs are on page 663.
	OrEqualBits(GPIO_PORT_B_DIR, 1<<5);

	// This is the "GPIO Alternate Function Select (GPIOAFSEL)" register.  See docs on page 0x420.  Setting
	// these bits to zero says to just use the pins as regular GPIO.
	ClearBits(GPIO_PORT_B_AFSEL, 1<<5);

	// This is the "GPIO Digital Enable (GPIODEN)" register.  Docs on page 682.  Here we're just enabling
	// PB5 to operate in digital mode.
	OrEqualBits(GPIO_PORT_B_DEN, 1<<5);
}

void SetupSSI2()
{
	// See pg 965, section 15.4 "Initialization and Configuration" for the steps to enable and initialize
	// SSI.  These steps are listed below by number:

	// 1. Enable the SSI module using the RCGCSSI register (see page 346).
	SetBits(SYSTEM_CONTROL_RCGCSSI, 1<<2);

	// 2. Enable the clock to the appropriate GPIO module via the RCGCGPIO register (see page 340). To find
	// out which GPIO port to enable, refer to Table 23-5 on page 1351.
	SetBits(SYSTEM_CONTROL_RCGCGPIO, 1<<1);

	ConfigureGPIOPinB5();

	// 3. Set the GPIO AFSEL bits for the appropriate pins (see page 671). To determine which GPIOs to
	// configure, see Table 23-4 on page 1344.
	// We're configuring pins 4 (1<<4), 6 (1<<6) and 7 (1<<7) here.
	OrEqualBits(GPIO_PORT_B_GPIOAFSEL, (1<<4) | (1<<6) | (1<<7));

	// 4. Configure the PMCn fields in the GPIOPCTL register to assign the SSI signals to the appropriate
	// pins. See page 688 and Table 23-5 on page 1351.
	OrEqualBits(GPIO_PORT_B_GPIOPCTL, (2<<16) | (2<<24) | (2<<28));

	// 5. Program the GPIODEN register to enable the pin's digital function. In addition, the drive
	// strength, drain select and pull-up/pull-down functions must be configured. Refer to
	// General-Purpose Input/Outputs (GPIOs) on page 649 for more information.
	OrEqualBits(GPIO_PORT_B_GPIODEN, (1<<4) | (1<<6) | (1<<7));

	// Here we disable the SSI module while configuring it.  See pg 972.
	ClearBits(SSI2_CR1, 1<<1);

	// Clearing bit 0 = "Normal serial port operation enabled"
	// Clearing bit 2 = Master mode
	// Clearing bit 4 = "The TXRIS interrupt indicates that the transmit FIFO is half full or less".
	// 0001.0101 binary = 0x15
	ClearBits(SSI2_CR1, 0x15);

	// Set the clock source to be the system clock.  See pg 984.
	SetBits(SSI2_CC, (0x00));

	// Set the clock prescale divisor to ten.  See pg 976.
	SetBits(SSI2_CPSR, 10);

	// Set the data transimission size to 8 bit data.  See pg 970.
	SetBits(SSI2_CR0, 0x7);

	// We're done configuring the SSI module so enable it.  See pg 972.
	OrEqualBits(SSI2_CR1, 1<<1);
}

void WriteSPI(uint8_t data)
{
	// See page 973 for details on the SSI Data register.  The location to write data is in bits 0-16.
	SetBits(SSI2_DR, data);

	// See page 974 for details on the SSI Status register.  The docs say if bit 0 (0x01 = 0000.0001 binary)
	// is set, the FIFO is empty.  I interrupt the following to be 'wait until the data has been consumed'.
	WaitUntilBitSet(SSI2_SR, 0x1);
}

uint8_t ReadSPI()
{
	// See page 974 for details on the SSI Status register.  The 2nd bit (0x04 = 0000.0100 binary) will be
	// set when the receive FIFO is not empty (i.e. when data exists to read).
	WaitUntilBitSet(SSI2_SR, 0x04);

	// See page 973 for details on the SSI Data register - we're just reading the lower 8 bits here.
	return ReadByte(SSI2_DR);
}

void Delay(unsigned int count)
{
	uint32_t i = 0;
	for(i = 0; i < count; i++);
}

// See pages 15 and 16 of the Microchip MCP23S17 datasheet
void SendCommandToMCP23S17(uint8_t control, uint8_t registerAddress, uint8_t deviceOpCode)
{
	SetBits(GPIO_PORT_B_DATA, 1<<5);   // Set CS high
	ClearBits(GPIO_PORT_B_DATA, 1<<5); // Set CS low to indicate a command is coming

	WriteSPI(MCP23S17_FIXED_BITS | control);
	WriteSPI(registerAddress);
	WriteSPI(deviceOpCode);

	// The following is a necessary delay after sending data.  The SSI communication waits very briefly
	// between sending data and then changing the clock signal to indicate we're done sending data.  If
	// we manually write a bit immediately there won't be enough delay and the communication won't work.
	// See https://www.youtube.com/watch?v=VprofC5ssi4 for more explanation.
	Delay(15);

	// Set the CS back to high
	SetBits(GPIO_PORT_B_DATA, 1<<5);
}

uint8_t ReadMCP23S17Pins(uint8_t control, uint8_t registerAddress)
{
	uint8_t returnValue = 0;

	SetBits(GPIO_PORT_B_DATA, 1<<5);   // Set CS high
	ClearBits(GPIO_PORT_B_DATA, 1<<5); // Set CS low to indicate a command is coming

	WriteSPI(MCP23S17_FIXED_BITS | control | MCP23S17_RWBIT_READ);
	WriteSPI(registerAddress);
	WriteSPI(0);  // Don't need to write anything to the third byte

	returnValue = ReadSPI();

	// The following is a necessary delay after sending data.  The SSI communication waits very briefly
	// between sending data and then changing the clock signal to indicate we're done sending data.  If
	// we manually write a bit immediately there won't be enough delay and the communication won't work.
	// See https://www.youtube.com/watch?v=VprofC5ssi4 for more explanation.
	Delay(15);

	// Set the CS back to high
	SetBits(GPIO_PORT_B_DATA, 1<<5);

	return returnValue;
}

void SetupMCP23S17Chips()
{
	// Since we have not yet enabled addressing (HAEN - what we're doing now) this first command will go to both chips.
	SendCommandToMCP23S17(0, MCP23S17_IOCON_REG, MCP23S17_HAEN);

	// Setup all pins of both MCP23S17 chips to be inputs
	SendCommandToMCP23S17(MCP23S17_CHIP1_ADDRESS | MCP23S17_RWBIT_READ, MCP23S17_IO_DIR_A_REG, MCP23S17_ALL_BITS_INPUT);
	SendCommandToMCP23S17(MCP23S17_CHIP1_ADDRESS | MCP23S17_RWBIT_READ, MCP23S17_IO_DIR_B_REG, MCP23S17_ALL_BITS_INPUT);
	SendCommandToMCP23S17(MCP23S17_CHIP2_ADDRESS | MCP23S17_RWBIT_READ, MCP23S17_IO_DIR_A_REG, MCP23S17_ALL_BITS_INPUT);
	SendCommandToMCP23S17(MCP23S17_CHIP2_ADDRESS | MCP23S17_RWBIT_READ, MCP23S17_IO_DIR_B_REG, MCP23S17_ALL_BITS_INPUT);
}

void main()
{
	// I'm making these 16 bit types instead of 8 bit since an unsigned 8 bit is dispalyed as a character instead of a
	// number in the Watch Window of Code Composer.
	uint16_t chip1GPA = 0;
	uint16_t chip1GPB = 0;
	uint16_t chip2GPA = 0;
	uint16_t chip2GPB = 0;

	SetupSSI2();

	SetupMCP23S17Chips();

	// Continually read the values of the GPA and GPB pins of both chips
	while(1)
	{
		chip1GPA = ReadMCP23S17Pins(MCP23S17_CHIP1_ADDRESS, MCP23S17_GPIO_A_DATA_REG);
		chip1GPB = ReadMCP23S17Pins(MCP23S17_CHIP1_ADDRESS, MCP23S17_GPIO_B_DATA_REG);
		chip2GPA = ReadMCP23S17Pins(MCP23S17_CHIP2_ADDRESS, MCP23S17_GPIO_A_DATA_REG);
		chip2GPB = ReadMCP23S17Pins(MCP23S17_CHIP2_ADDRESS, MCP23S17_GPIO_B_DATA_REG);
	}
}
